/**************************************************************/
/*                                                            */
/* Hardihosh.c                                                */
/*                                                            */
/* Generates quasirandom output text based on the input text  */
/* Feel free to play with it, but please leave my name on it  */
/* somewhere, and don't use it in anything commercial without */
/* asking first. (altho I can't imagine why you'd want to...) */
/* Also, original tabbing was 4 spaces, so if you use 8 some  */
/* stuff is going to seem reeeaally indented                  */
/*                                                            */
/*                                                            */
/* Copyleft Ryan Schmidt 1998                                 */
/* Ryan Schmidt (dgawd@shockware.com)                         */
/* web.shockware.com/users/dgawd                              */
/*                                                            */
/**************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define MAXLOOP 6   /* maximum number of words kept to avoid looping */

/* word linked-list structure */
typedef struct _WORDSTRUCT WORDSTRUCT;
typedef WORDSTRUCT *WORD;
struct _WORDSTRUCT {
	char	*string;
	int	prob;
	WORD follow;
	WORD next;
};

/* wordfile structure */
struct _WORDLIST {
	int length;
	WORD top;
};
typedef struct _WORDLIST WORDLIST;

int stoploop(char *);
int readwords(char *, WORDLIST *);
char *wtok(FILE *);
int isourpunct(char);
WORD addword(WORD *, char *);
WORD searchwordlist(WORD, char*);
void dumpwordlist(WORD);
void fdumpwordlist(WORD);
WORD pick(WORD);
int printwords(FILE *, WORDLIST *);



int main(int argc, char *argv[])
{

	WORDLIST wlist;
	FILE *fp;

	if(argc < 2) {
		printf("Usage: hardihosh <infile>\n");
		printf("Options: hardihosh <infile> <outfile>\n");
		exit(1);
	}	
	
	if( (readwords(argv[1], &wlist)) < 0) {
		printf("Error reading infile - aborting\n");
		exit(1);
	}
	
	if(argc == 3) {
		if( (fp = fopen(argv[2],"w")) == NULL) {
			perror(argv[2]);
			goto abort;
		}
		if(printwords(fp,&wlist) < 0) {
			perror(argv[0]);
			goto abort;
		}
	}else if(printwords(stdout, &wlist) < 0) {
			perror(argv[0]);
			goto abort;
		}

	printf("done...\n");	
	return(1);

abort:
		printf("Aborting...\n");
		exit(-1);
}

/* prevents text loops in the output file                        */
/* this is sort of a hack - it would be better done by stripping */
/* loops out of the list, but this seems to  be faster           */
int stoploop(char *string)
{

	static char *lstring[MAXLOOP];
	static int tolerance = MAXLOOP;
	static int lptr = 0;
	static int init = 0;
	int i;
	int stolerance;

	if(string == NULL)
		return(1);
	if(init == 0) {                       /* one-time initialization */
		for(i = 0; i < MAXLOOP; i++)
			lstring[i] = NULL;
		init = 1;
	}

	stolerance = tolerance;
	for(i = 0; i < MAXLOOP && lstring[i] != NULL; i++){ 
		if(strcmp(lstring[i],string) == 0)  /* scan loop string for matches */
			stolerance--;
	}
	if(strcmp(string,".") == 0)
		tolerance = MAXLOOP;
	else if(stolerance == tolerance)
		tolerance = (tolerance == MAXLOOP) ? MAXLOOP : tolerance++;
	else if(stolerance != tolerance)
		tolerance--;
	if(tolerance < 0 ) {                 /* if too many repeats, nuke list */
		tolerance = MAXLOOP;              /* and return -1 to indicate loop */
		for(i = 0; i < MAXLOOP; i++){
			free(lstring[i]);
			lstring[i] = NULL;
		}
		lptr = 0;
		return(-1);
	}

	if(lptr + 1 >= MAXLOOP){            /* wrap loop string index */
		lptr = 0;
	}else
		lptr++;

	free(lstring[i]);
	lstring[lptr] = malloc(strlen(string) + 1);
	strcpy(lstring[lptr],string);
	
	return(1);
}


/* generates and prints new text to text file */
int printwords(FILE *fp, WORDLIST *wordlist)
{
	
	WORD search, selected;
	int i, done;

	srand(time(NULL));

	search = searchwordlist(wordlist->top, ".");   /* start at a proper first word */
	i = wordlist->length;

	done = 0;	                     /* this makes sure we end with a period */
	while( (i-- > 0 || done == 1) && i > -50) {
		if(i == 0)
			done = 1;
		if(search == NULL)
			return(-1);
		if( (stoploop(search->string) < 0) || (search->follow == NULL)) {  
			fprintf(fp,". ");      /* deal with no follow or loop */
			search = searchwordlist(wordlist->top,".");
			continue;
		}
		selected = pick(search);      /* pick one of the follow words */
		if(selected == NULL)
			return(-1);                /* start a new paragraph */
		if(strcmp(selected->string,"_START_NEW_PARAGRAPH") == 0) 
			fprintf(fp,"\n\n\t");
		else {
			if(isourpunct(selected->string[0])){    /* erase space before .'s */
				fprintf(fp,"\b%s ",selected->string);
				if(done == 1)
					done = 0;
			}else
				fprintf(fp,"%s ",selected->string);
		}                             /* find next word node... */
		search = searchwordlist(wordlist->top,selected->string);
	}
	return(1);
}


/* picks a random word from the follow list of the word node you pass it */
WORD pick(WORD plist)
{
	
	int ptotal, prand, wltotal, ntotal;
	WORD current;

	current = plist->follow;	
	ptotal = 0;
	while(current != NULL){      /* get prob total */
		ptotal += current->prob;
		current = current->next;
	}                            /* pick a word */
	prand = (int)((float)ptotal * rand() / (RAND_MAX + 1.0));

	current = plist->follow;
	wltotal = ntotal = 0;
	while(wltotal < prand) {        /* find word */
		if(ntotal == 0)
			ntotal = current->prob;
		else 
			if(--ntotal == 0)
				current = current->next;
		wltotal++;
	}
	return(current); 
}

/* read the words from a text file the into wordlist and accompanying */
/* follow-wordlists in the wlist structure                            */
int readwords(char *path, WORDLIST *wlist)
{
	FILE *fp;
	char *nextword;
	WORD last;
	
	wlist->length = 0;
	wlist->top = NULL;
	last = NULL;
	
	if( (fp = fopen(path,"r")) == NULL)
		return(-1);

	while( (nextword = wtok(fp)) != NULL){
		if(last != NULL && strcmp(nextword,last->string) != 0 )
			addword(&last->follow,nextword);
		if( (last = addword(&wlist->top, nextword)) != NULL) 
			wlist->length++;
		else
			return(-1);
		free(nextword);
	}	
	if(nextword != NULL)
		addword(&last->follow,nextword);

	fclose(fp);
	return(1);
	
}

  
/* add a unique word to word list, return new node or the node it */
/* matches if word is not unique                                  */
/* currently doesn't check malloc() returns - will probably crash */
/* if malloc() fails                                              */
WORD addword(WORD *top, char *wstring)
{
	WORD new, current;

	if(*top == NULL){     /* empty list - adding first word*/
		*top = (WORD)malloc(sizeof(WORDSTRUCT));
		(*top)->string = (char *)malloc(strlen(wstring) + 1);
		strcpy((*top)->string,wstring);
		(*top)->prob = 1;
		(*top)->follow = NULL;
		(*top)->next = NULL;
		return(*top);
	}
	
	if( (current = searchwordlist(*top,wstring)) == NULL)
	{
		current = *top;
		while(current->next != NULL)
			current = current->next;              /* add new wordstruct to */
		new = (WORD)malloc(sizeof(WORDSTRUCT));  /* end of our list */  
		new->string = (char *)malloc(strlen(wstring) + 1);
		strcpy(new->string,wstring);
		new->prob = 1;
		new->follow = NULL;
		new->next = NULL;
		current->next = new;
		return(new);
	}else{
		current->prob++;
		return(current);                 /* search matched word, return it */
	}

	return(NULL);                       /* we should never get here */
}


/* search through word list for word, return node it's in */
WORD searchwordlist(WORD top, char *wstring)
{
	WORD current;

	current = top;	
	while(current != NULL && strcmp(current->string,wstring) != 0)
		current = current->next;

	return(current);
}


/* print out all the words in our list (debugging */
void dumpwordlist(WORD top)
{
	WORD current;
	
	current = top;
	while(current != NULL) {
		printf("%s - p%d",current->string,current->prob);
		fdumpwordlist(current->follow);
		printf("\n");
		current = current->next;
	}
}

void fdumpwordlist(WORD top)
{
	WORD current;
	
	current = top;
	while(current != NULL) {
		printf(" %s (p%d) ->",current->string,current->prob);
		current = current->next;
	}
	printf("NULL\n");
}

/* get next word, ./?/, punctuation character, or _START_NEW_PARAGRAPH */
char *wtok(FILE *in)
{
	
	char *nw;
	int i, len;
	char c;
	
	nw = (char *)malloc(64);
	if(nw == NULL)
		return(NULL);
	len = 64;
	i = 0;

repeat:

	c = fgetc(in);
	while( c == ' ' || c == '\t') /* skip whitespace */
		c = fgetc(in);
                                 /* read a word... */
	while(c != ' ' && !isourpunct(c) && c != EOF){
		if(i == len - 1) {         /* get more memory if we run out */
			realloc(nw, len + 64);
			if(nw == NULL)
				return(NULL);
			len += 64;
		}
			nw[i++] = c;
			c = fgetc(in);
	}
	if(i > 0) {
		nw[i] = '\0';
		if(isourpunct(c))
			ungetc(c, in);
	}else{
		if(isourpunct(c)) {
			if(c == '\n') {    /* test to see if we are at new paragraph, else */
				c = fgetc(in);  /* read next word */
				if( c == '\t' || c == '\n') 
					sprintf(nw,"_START_NEW_PARAGRAPH");
				else {
					ungetc(c,in);
					goto repeat;
				}
			} else
				sprintf(nw,"%c",c);
		} else {
			if(c == ' ')        
				goto repeat;     /* keep reading */
			free(nw);
			nw = NULL;
		}
	}

	return(nw);
} /* END WTOK */


int isourpunct(char c)
{
	if(c == ',' || c == '.' || c == '?' || c == '!' || c == '\n')
		return(1);
	else
		return(0);
}